# -*- coding: utf-8 -*-
import asyncio

import blivedm
import blivedm.models.open_live as open_models
import blivedm.models.web as web_models

# 在开放平台申请的开发者密钥
ACCESS_KEY_ID = ''
ACCESS_KEY_SECRET = ''
# 在开放平台创建的项目ID
APP_ID = 0
# 主播身份码
ROOM_OWNER_AUTH_CODE = ''


async def main():
    await run_single_client()


async def run_single_client():
    """
    演示监听一个直播间
    """
    client = blivedm.OpenLiveClient(
        access_key_id=ACCESS_KEY_ID,
        access_key_secret=ACCESS_KEY_SECRET,
        app_id=APP_ID,
        room_owner_auth_code=ROOM_OWNER_AUTH_CODE,
    )
    handler = MyHandler()
    client.set_handler(handler)

    client.start()
    try:
        # 演示70秒后停止
        # await asyncio.sleep(70)
        # client.stop()

        await client.join()
    finally:
        await client.stop_and_close()


class MyHandler(blivedm.BaseHandler):
    def _on_heartbeat(self, client: blivedm.BLiveClient, message: web_models.HeartbeatMessage):
        print(f'[{client.room_id}] 心跳')

    def _on_open_live_danmaku(self, client: blivedm.OpenLiveClient, message: open_models.DanmakuMessage):
        print(f'[{message.room_id}] {message.uname}：{message.msg}')

    def _on_open_live_gift(self, client: blivedm.OpenLiveClient, message: open_models.GiftMessage):
        coin_type = '金瓜子' if message.paid else '银瓜子'
        total_coin = message.price * message.gift_num
        print(f'[{message.room_id}] {message.uname} 赠送{message.gift_name}x{message.gift_num}'
              f' （{coin_type}x{total_coin}）')

    def _on_open_live_buy_guard(self, client: blivedm.OpenLiveClient, message: open_models.GuardBuyMessage):
        print(f'[{message.room_id}] {message.user_info.uname} 购买 大航海等级={message.guard_level}')

    def _on_open_live_super_chat(
        self, client: blivedm.OpenLiveClient, message: open_models.SuperChatMessage
    ):
        print(f'[{message.room_id}] 醒目留言 ¥{message.rmb} {message.uname}：{message.message}')

    def _on_open_live_super_chat_delete(
        self, client: blivedm.OpenLiveClient, message: open_models.SuperChatDeleteMessage
    ):
        print(f'[{message.room_id}] 删除醒目留言 message_ids={message.message_ids}')

    def _on_open_live_like(self, client: blivedm.OpenLiveClient, message: open_models.LikeMessage):
        print(f'[{message.room_id}] {message.uname} 点赞')


if __name__ == '__main__':
    asyncio.run(main())



# -*- coding: utf-8 -*-
import asyncio
import enum
import json
import logging
import struct
import zlib
from typing import Callable, NamedTuple, Optional, Union

import aiohttp
import brotli

from .. import handlers, utils

logger = logging.getLogger("blivedm")

USER_AGENT = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/102.0.0.0 Safari/537.36"

HEADER_STRUCT = struct.Struct(">I2H2I")


class HeaderTuple(NamedTuple):
    pack_len: int
    raw_header_size: int
    ver: int
    operation: int
    seq_id: int


# WS_BODY_PROTOCOL_VERSION
class ProtoVer(enum.IntEnum):
    NORMAL = 0
    HEARTBEAT = 1
    DEFLATE = 2
    BROTLI = 3


# go-common\app\service\main\broadcast\model\operation.go
class Operation(enum.IntEnum):
    HANDSHAKE = 0
    HANDSHAKE_REPLY = 1
    HEARTBEAT = 2
    HEARTBEAT_REPLY = 3
    SEND_MSG = 4
    SEND_MSG_REPLY = 5
    DISCONNECT_REPLY = 6
    AUTH = 7
    AUTH_REPLY = 8
    RAW = 9
    PROTO_READY = 10
    PROTO_FINISH = 11
    CHANGE_ROOM = 12
    CHANGE_ROOM_REPLY = 13
    REGISTER = 14
    REGISTER_REPLY = 15
    UNREGISTER = 16
    UNREGISTER_REPLY = 17
    # B站业务自定义OP
    # MinBusinessOp = 1000
    # MaxBusinessOp = 10000


# WS_AUTH
class AuthReplyCode(enum.IntEnum):
    OK = 0
    TOKEN_ERROR = -101


class InitError(Exception):
    """初始化失败"""


class AuthError(Exception):
    """认证失败"""


DEFAULT_RECONNECT_POLICY = utils.make_constant_retry_policy(1)


class WebSocketClientBase:
    """
    基于WebSocket的客户端

    :param session: cookie、连接池
    :param heartbeat_interval: 发送心跳包的间隔时间（秒）
    """

    def __init__(
        self,
        session: Optional[aiohttp.ClientSession] = None,
        heartbeat_interval: float = 30,
    ):
        if session is None:
            self._session = aiohttp.ClientSession(
                timeout=aiohttp.ClientTimeout(total=10)
            )
            self._own_session = True
        else:
            self._session = session
            self._own_session = False
            assert self._session.loop is asyncio.get_event_loop()  # noqa

        self._heartbeat_interval = heartbeat_interval

        self._need_init_room = True
        self._handler: Optional[handlers.HandlerInterface] = None
        """消息处理器"""
        self._get_reconnect_interval: Callable[
            [int, int], float
        ] = DEFAULT_RECONNECT_POLICY
        """重连间隔时间增长策略"""

        # 在调用init_room后初始化的字段
        self._room_id: Optional[int] = None

        # 在运行时初始化的字段
        self._websocket: Optional[aiohttp.ClientWebSocketResponse] = None
        """WebSocket连接"""
        self._network_future: Optional[asyncio.Future] = None
        """网络协程的future"""
        self._heartbeat_timer_handle: Optional[asyncio.TimerHandle] = None
        """发心跳包定时器的handle"""

    @property
    def is_running(self) -> bool:
        """
        本客户端正在运行，注意调用stop后还没完全停止也算正在运行
        """
        return self._network_future is not None

    @property
    def room_id(self) -> Optional[int]:
        """
        房间ID，调用init_room后初始化
        """
        return self._room_id

    def set_handler(self, handler: Optional["handlers.HandlerInterface"]):
        """
        设置消息处理器

        注意消息处理器和网络协程运行在同一个协程，如果处理消息耗时太长会阻塞接收消息。如果是CPU密集型的任务，建议将消息推到线程池处理；
        如果是IO密集型的任务，应该使用async函数，并且在handler里使用create_task创建新的协程

        :param handler: 消息处理器
        """
        self._handler = handler

    def set_reconnect_policy(self, get_reconnect_interval: Callable[[int, int], float]):
        """
        设置重连间隔时间增长策略

        :param get_reconnect_interval: 一个可调用对象，输入重试次数 (retry_count, total_retry_count)，返回间隔时间
        """
        self._get_reconnect_interval = get_reconnect_interval

    def start(self):
        """
        启动本客户端
        """
        if self.is_running:
            logger.warning(
                "room=%s client is running, cannot start() again", self.room_id
            )
            return

        self._network_future = asyncio.create_task(self._network_coroutine_wrapper())

    def stop(self):
        """
        停止本客户端
        """
        if not self.is_running:
            logger.warning(
                "room=%s client is stopped, cannot stop() again", self.room_id
            )
            return

        if self._network_future is not None:
            self._network_future.cancel()

    async def stop_and_close(self):
        """
        便利函数，停止本客户端并释放本客户端的资源，调用后本客户端将不可用
        """
        if self.is_running:
            self.stop()
            await self.join()
        await self.close()

    async def join(self):
        """
        等待本客户端停止
        """
        if not self.is_running:
            logger.warning("room=%s client is stopped, cannot join()", self.room_id)
            return

        if self._network_future is not None:
            await asyncio.shield(self._network_future)

    async def close(self):
        """
        释放本客户端的资源，调用后本客户端将不可用
        """
        if self.is_running:
            logger.warning(
                "room=%s is calling close(), but client is running", self.room_id
            )

        # 如果session是自己创建的则关闭session
        if self._own_session:
            await self._session.close()

    async def init_room(self) -> bool:
        """
        初始化连接房间需要的字段

        :return: True代表没有降级，如果需要降级后还可用，重载这个函数返回True
        """
        raise NotImplementedError

    @staticmethod
    def _make_packet(data: Union[dict, str, bytes], operation: int) -> bytes:
        """
        创建一个要发送给服务器的包

        :param data: 包体JSON数据
        :param operation: 操作码，见Operation
        :return: 整个包的数据
        """
        if isinstance(data, dict):
            body = json.dumps(data).encode("utf-8")
        elif isinstance(data, str):
            body = data.encode("utf-8")
        else:
            body = data
        header = HEADER_STRUCT.pack(
            *HeaderTuple(
                pack_len=HEADER_STRUCT.size + len(body),
                raw_header_size=HEADER_STRUCT.size,
                ver=1,
                operation=operation,
                seq_id=1,
            )
        )
        return header + body

    async def _network_coroutine_wrapper(self):
        """
        负责处理网络协程的异常，网络协程具体逻辑在_network_coroutine里
        """
        exc = None
        try:
            await self._network_coroutine()
        except asyncio.CancelledError:
            # 正常停止
            pass
        except Exception as e:
            logger.exception(
                "room=%s _network_coroutine() finished with exception:", self.room_id
            )
            exc = e
        finally:
            logger.debug("room=%s _network_coroutine() finished", self.room_id)
            self._network_future = None

        if self._handler is not None:
            self._handler.on_client_stopped(self, exc)

    async def _network_coroutine(self):
        """
        网络协程，负责连接服务器、接收消息、解包
        """
        # retry_count在连接成功后会重置为0，total_retry_count不会
        retry_count = 0
        total_retry_count = 0
        while True:
            try:
                await self._on_before_ws_connect(retry_count)

                # 连接
                async with self._session.ws_connect(
                    self._get_ws_url(retry_count),
                    headers={"User-Agent": utils.USER_AGENT},  # web端的token也会签名UA
                    receive_timeout=self._heartbeat_interval + 5,
                ) as websocket:
                    self._websocket = websocket
                    await self._on_ws_connect()

                    # 处理消息
                    message: aiohttp.WSMessage
                    async for message in websocket:
                        await self._on_ws_message(message)
                        # 至少成功处理1条消息
                        retry_count = 0

            except (aiohttp.ClientConnectionError, asyncio.TimeoutError):
                # 掉线重连
                pass
            except AuthError:
                # 认证失败了，应该重新获取token再重连
                logger.exception(
                    "room=%d auth failed, trying init_room() again", self.room_id
                )
                self._need_init_room = True
            finally:
                self._websocket = None
                await self._on_ws_close()

            # 准备重连
            retry_count += 1
            total_retry_count += 1
            logger.warning(
                "room=%d is reconnecting, retry_count=%d, total_retry_count=%d",
                self.room_id,
                retry_count,
                total_retry_count,
            )
            await asyncio.sleep(
                self._get_reconnect_interval(retry_count, total_retry_count)
            )

    async def _on_before_ws_connect(self, retry_count):
        """
        在每次建立连接之前调用，可以用来初始化房间
        """
        if not self._need_init_room:
            return

        if not await self.init_room():
            raise InitError("init_room() failed")
        self._need_init_room = False

    def _get_ws_url(self, retry_count) -> str:
        """
        返回WebSocket连接的URL，可以在这里做故障转移和负载均衡
        """
        raise NotImplementedError

    async def _on_ws_connect(self):
        """
        WebSocket连接成功
        """
        await self._send_auth()
        self._heartbeat_timer_handle = asyncio.get_running_loop().call_later(
            self._heartbeat_interval, self._on_send_heartbeat
        )

    async def _on_ws_close(self):
        """
        WebSocket连接断开
        """
        if self._heartbeat_timer_handle is not None:
            self._heartbeat_timer_handle.cancel()
            self._heartbeat_timer_handle = None

    async def _send_auth(self):
        """
        发送认证包
        """
        raise NotImplementedError

    def _on_send_heartbeat(self):
        """
        定时发送心跳包的回调
        """
        if self._websocket is None or self._websocket.closed:
            self._heartbeat_timer_handle = None
            return

        self._heartbeat_timer_handle = asyncio.get_running_loop().call_later(
            self._heartbeat_interval, self._on_send_heartbeat
        )
        asyncio.create_task(self._send_heartbeat())

    async def _send_heartbeat(self):
        """
        发送心跳包
        """
        if self._websocket is None or self._websocket.closed:
            return

        try:
            await self._websocket.send_bytes(self._make_packet({}, Operation.HEARTBEAT))
        except (ConnectionResetError, aiohttp.ClientConnectionError) as e:
            logger.warning("room=%d _send_heartbeat() failed: %r", self.room_id, e)
        except Exception:  # noqa
            logger.exception("room=%d _send_heartbeat() failed:", self.room_id)

    async def _on_ws_message(self, message: aiohttp.WSMessage):
        """
        收到WebSocket消息

        :param message: WebSocket消息
        """
        if message.type != aiohttp.WSMsgType.BINARY:
            logger.warning(
                "room=%d unknown websocket message type=%s, data=%s",
                self.room_id,
                message.type,
                message.data,
            )
            return

        try:
            await self._parse_ws_message(message.data)
        except AuthError:
            # 认证失败，让外层处理
            raise
        except Exception:  # noqa
            logger.exception("room=%d _parse_ws_message() error:", self.room_id)

    async def _parse_ws_message(self, data: bytes):
        """
        解析WebSocket消息

        :param data: WebSocket消息数据
        """
        offset = 0
        try:
            header = HeaderTuple(*HEADER_STRUCT.unpack_from(data, offset))
        except struct.error:
            logger.exception(
                "room=%d parsing header failed, offset=%d, data=%s",
                self.room_id,
                offset,
                data,
            )
            return

        if header.operation in (Operation.SEND_MSG_REPLY, Operation.AUTH_REPLY):
            # 业务消息，可能有多个包一起发，需要分包
            while True:
                body = data[offset + header.raw_header_size : offset + header.pack_len]
                await self._parse_business_message(header, body)

                offset += header.pack_len
                if offset >= len(data):
                    break

                try:
                    header = HeaderTuple(*HEADER_STRUCT.unpack_from(data, offset))
                except struct.error:
                    logger.exception(
                        "room=%d parsing header failed, offset=%d, data=%s",
                        self.room_id,
                        offset,
                        data,
                    )
                    break

        elif header.operation == Operation.HEARTBEAT_REPLY:
            # 服务器心跳包，前4字节是人气值，后面是客户端发的心跳包内容
            # pack_len不包括客户端发的心跳包内容，不知道是不是服务器BUG
            body = data[
                offset + header.raw_header_size : offset + header.raw_header_size + 4
            ]
            popularity = int.from_bytes(body, "big")
            # 自己造个消息当成业务消息处理
            body = {"cmd": "_HEARTBEAT", "data": {"popularity": popularity}}
            self._handle_command(body)

        else:
            # 未知消息
            body = data[offset + header.raw_header_size : offset + header.pack_len]
            logger.warning(
                "room=%d unknown message operation=%d, header=%s, body=%s",
                self.room_id,
                header.operation,
                header,
                body,
            )

    async def _parse_business_message(self, header: HeaderTuple, body: bytes):
        """
        解析业务消息
        """
        if header.operation == Operation.SEND_MSG_REPLY:
            # 业务消息
            if header.ver == ProtoVer.BROTLI:
                # 压缩过的先解压，为了避免阻塞网络线程，放在其他线程执行
                body = await asyncio.get_running_loop().run_in_executor(
                    None, brotli.decompress, body
                )
                await self._parse_ws_message(body)
            elif header.ver == ProtoVer.DEFLATE:
                # web端已经不用zlib压缩了，但是开放平台会用
                body = await asyncio.get_running_loop().run_in_executor(
                    None, zlib.decompress, body
                )
                await self._parse_ws_message(body)
            elif header.ver == ProtoVer.NORMAL:
                # 没压缩过的直接反序列化，因为有万恶的GIL，这里不能并行避免阻塞
                if len(body) != 0:
                    try:
                        body = json.loads(body.decode("utf-8"))
                        self._handle_command(body)
                    except Exception:
                        logger.error("room=%d, body=%s", self.room_id, body)
                        raise
            else:
                # 未知格式
                logger.warning(
                    "room=%d unknown protocol version=%d, header=%s, body=%s",
                    self.room_id,
                    header.ver,
                    header,
                    body,
                )

        elif header.operation == Operation.AUTH_REPLY:
            # 认证响应
            body = json.loads(body.decode("utf-8"))
            print(f"Body: {body}, Type: {type(body)}")
            if body["code"] != AuthReplyCode.OK:
                raise AuthError(f"auth reply error, code={body['code']}, body={body}")
            if self._websocket is not None:
                await self._websocket.send_bytes(
                    self._make_packet({}, Operation.HEARTBEAT)
                )

        else:
            # 未知消息
            logger.warning(
                "room=%d unknown message operation=%d, header=%s, body=%s",
                self.room_id,
                header.operation,
                header,
                body,
            )

    def _handle_command(self, command: dict):
        """
        处理业务消息

        :param command: 业务消息
        """
        if self._handler is None:
            return
        try:
            # 为什么不做成异步的：
            # 1. 为了保持处理消息的顺序，这里不使用call_soon、create_task等方法延迟处理
            # 2. 如果支持handle使用async函数，用户可能会在里面处理耗时很长的异步操作，导致网络协程阻塞
            # 这里做成同步的，强制用户使用create_task或消息队列处理异步操作，这样就不会阻塞网络协程
            self._handler.handle(self, command)
        except Exception as e:
            logger.exception(
                "room=%d _handle_command() failed, command=%s",
                self.room_id,
                command,
                exc_info=e,
            )



# -*- coding: utf-8 -*-
import asyncio
import datetime
import hashlib
import hmac
import json
import logging
import uuid
from typing import List, Optional

import aiohttp

from . import ws_base

__all__ = ("OpenLiveClient",)

logger = logging.getLogger("blivedm")

START_URL = "https://live-open.biliapi.com/v2/app/start"
HEARTBEAT_URL = "https://live-open.biliapi.com/v2/app/heartbeat"
END_URL = "https://live-open.biliapi.com/v2/app/end"


class OpenLiveClient(ws_base.WebSocketClientBase):
    """
    开放平台客户端

    文档参考：https://open-live.bilibili.com/document/

    :param access_key_id: 在开放平台申请的access_key_id
    :param access_key_secret: 在开放平台申请的access_key_secret
    :param app_id: 在开放平台创建的项目ID
    :param room_owner_auth_code: 主播身份码
    :param session: cookie、连接池
    :param heartbeat_interval: 发送连接心跳包的间隔时间（秒）
    :param game_heartbeat_interval: 发送项目心跳包的间隔时间（秒）
    """

    def __init__(
        self,
        access_key_id: str,
        access_key_secret: str,
        app_id: int,
        room_owner_auth_code: str,
        *,
        session: Optional[aiohttp.ClientSession] = None,
        heartbeat_interval=30,
        game_heartbeat_interval=20,
    ):
        super().__init__(session, heartbeat_interval)

        self._access_key_id = access_key_id
        self._access_key_secret = access_key_secret
        self._app_id = app_id
        self._room_owner_auth_code = room_owner_auth_code
        self._game_heartbeat_interval = game_heartbeat_interval

        # 在调用init_room后初始化的字段
        self._room_owner_uid: Optional[int] = None
        """主播用户ID"""
        self._room_owner_open_id: Optional[str] = None
        """主播Open ID"""
        self._host_server_url_list: Optional[List[str]] = []
        """弹幕服务器URL列表"""
        self._auth_body: Optional[str] = None
        """连接弹幕服务器用的认证包内容"""
        self._game_id: Optional[str] = None
        """项目场次ID"""

        # 在运行时初始化的字段
        self._game_heartbeat_timer_handle: Optional[asyncio.TimerHandle] = None
        """发项目心跳包定时器的handle"""

    @property
    def room_owner_uid(self) -> Optional[int]:
        """
        主播用户ID，调用init_room后初始化
        """
        return self._room_owner_uid

    @property
    def room_owner_open_id(self) -> Optional[str]:
        """
        主播Open ID，调用init_room后初始化
        """
        return self._room_owner_open_id

    @property
    def room_owner_auth_code(self):
        """
        主播身份码
        """
        return self._room_owner_auth_code

    @property
    def app_id(self):
        """
        在开放平台创建的项目ID
        """
        return self._app_id

    @property
    def game_id(self) -> Optional[str]:
        """
        项目场次ID，调用init_room后初始化
        """
        return self._game_id

    async def close(self):
        """
        释放本客户端的资源，调用后本客户端将不可用
        """
        if self.is_running:
            logger.warning(
                "room=%s is calling close(), but client is running", self.room_id
            )

        if self._game_heartbeat_timer_handle is not None:
            self._game_heartbeat_timer_handle.cancel()
            self._game_heartbeat_timer_handle = None
        await self._end_game()

        await super().close()

    def _request_open_live(self, url, body: dict):
        body_bytes = json.dumps(body).encode("utf-8")
        headers = {
            "x-bili-accesskeyid": self._access_key_id,
            "x-bili-content-md5": hashlib.md5(body_bytes).hexdigest(),
            "x-bili-signature-method": "HMAC-SHA256",
            "x-bili-signature-nonce": uuid.uuid4().hex,
            "x-bili-signature-version": "1.0",
            "x-bili-timestamp": str(int(datetime.datetime.now().timestamp())),
        }

        str_to_sign = "\n".join(f"{key}:{value}" for key, value in headers.items())
        signature = hmac.new(
            self._access_key_secret.encode("utf-8"),
            str_to_sign.encode("utf-8"),
            hashlib.sha256,
        ).hexdigest()
        headers["Authorization"] = signature

        headers["Content-Type"] = "application/json"
        headers["Accept"] = "application/json"
        return self._session.post(url, headers=headers, data=body_bytes)

    async def init_room(self):
        """
        开启项目，并初始化连接房间需要的字段

        :return: 是否成功
        """
        if not await self._start_game():
            return False

        if self._game_id != "" and self._game_heartbeat_timer_handle is None:
            self._game_heartbeat_timer_handle = asyncio.get_running_loop().call_later(
                self._game_heartbeat_interval, self._on_send_game_heartbeat
            )
        return True

    async def _start_game(self):
        try:
            async with self._request_open_live(
                START_URL, {"code": self._room_owner_auth_code, "app_id": self._app_id}
            ) as res:
                if res.status != 200:
                    logger.warning(
                        "_start_game() failed, status=%d, reason=%s",
                        res.status,
                        res.reason,
                    )
                    return False
                data = await res.json()
                if data["code"] != 0:
                    logger.warning(
                        "_start_game() failed, code=%d, message=%s, request_id=%s",
                        data["code"],
                        data["message"],
                        data["request_id"],
                    )
                    return False
                if not self._parse_start_game(data["data"]):
                    return False
        except (aiohttp.ClientConnectionError, asyncio.TimeoutError):
            logger.exception("_start_game() failed:")
            return False
        return True

    def _parse_start_game(self, data):
        self._game_id = data["game_info"]["game_id"]
        websocket_info = data["websocket_info"]
        self._auth_body = websocket_info["auth_body"]
        self._host_server_url_list = websocket_info["wss_link"]
        anchor_info = data["anchor_info"]
        self._room_id = anchor_info["room_id"]
        self._room_owner_uid = anchor_info["uid"]
        self._room_owner_open_id = anchor_info["open_id"]
        return True

    async def _end_game(self):
        """
        关闭项目。建议关闭客户端时保证调用到这个函数（close会调用），否则可能短时间内无法重复连接同一个房间
        """
        if self._game_id in (None, ""):
            return True

        try:
            async with self._request_open_live(
                END_URL, {"app_id": self._app_id, "game_id": self._game_id}
            ) as res:
                if res.status != 200:
                    logger.warning(
                        "room=%d _end_game() failed, status=%d, reason=%s",
                        self._room_id,
                        res.status,
                        res.reason,
                    )
                    return False
                data = await res.json()
                code = data["code"]
                if code != 0:
                    if code in (7000, 7003):
                        # 项目已经关闭了也算成功
                        return True

                    logger.warning(
                        "room=%d _end_game() failed, code=%d, message=%s, request_id=%s",
                        self._room_id,
                        code,
                        data["message"],
                        data["request_id"],
                    )
                    return False
        except (aiohttp.ClientConnectionError, asyncio.TimeoutError):
            logger.exception("room=%d _end_game() failed:", self._room_id)
            return False
        return True

    def _on_send_game_heartbeat(self):
        """
        定时发送项目心跳包的回调
        """
        self._game_heartbeat_timer_handle = asyncio.get_running_loop().call_later(
            self._game_heartbeat_interval, self._on_send_game_heartbeat
        )
        asyncio.create_task(self._send_game_heartbeat())

    async def _send_game_heartbeat(self):
        """
        发送项目心跳包
        """
        if self._game_id in (None, ""):
            logger.warning(
                "game=%d _send_game_heartbeat() failed, game_id not found",
                self._game_id,
            )
            return False

        try:
            # 保存一下，防止await之后game_id改变
            game_id = self._game_id
            async with self._request_open_live(
                HEARTBEAT_URL, {"game_id": game_id}
            ) as res:
                if res.status != 200:
                    logger.warning(
                        "room=%d _send_game_heartbeat() failed, status=%d, reason=%s",
                        self._room_id,
                        res.status,
                        res.reason,
                    )
                    return False
                data = await res.json()
                code = data["code"]
                if code != 0:
                    logger.warning(
                        "room=%d _send_game_heartbeat() failed, code=%d, message=%s, request_id=%s",
                        self._room_id,
                        code,
                        data["message"],
                        data["request_id"],
                    )

                    if code == 7003 and self._game_id == game_id:
                        # 项目异常关闭，可能是心跳超时，需要重新开启项目
                        self._need_init_room = True
                        if self._websocket is not None and not self._websocket.closed:
                            await self._websocket.close()

                    return False
        except (aiohttp.ClientConnectionError, asyncio.TimeoutError):
            logger.exception("room=%d _send_game_heartbeat() failed:", self._room_id)
            return False
        return True

    async def _on_before_ws_connect(self, retry_count):
        """
        在每次建立连接之前调用，可以用来初始化房间
        """
        # 重连次数太多则重新init_room，保险
        reinit_period = max(3, len(self._host_server_url_list or ()))
        if retry_count > 0 and retry_count % reinit_period == 0:
            self._need_init_room = True
        await super()._on_before_ws_connect(retry_count)

    def _get_ws_url(self, retry_count) -> str:
        """
        返回WebSocket连接的URL，可以在这里做故障转移和负载均衡
        """
        if self._host_server_url_list is None:
            # Handle the case where the list is None. For example:
            raise ValueError("Host server URL list is not initialized.")
        return self._host_server_url_list[retry_count % len(self._host_server_url_list)]

    async def _send_auth(self):
        """
        发送认证包
        """
        if self._websocket is not None and self._auth_body is not None:
            await self._websocket.send_bytes(
                self._make_packet(self._auth_body, ws_base.Operation.AUTH)
            )



# -*- coding: utf-8 -*-
import logging
from typing import Any, Callable, Dict, Optional

from .clients import ws_base
from .models import web as web_models, open_live as open_models

__all__ = (
    "HandlerInterface",
    "BaseHandler",
)

logger = logging.getLogger("blivedm")

logged_unknown_cmds = {
    "COMBO_SEND",
    "ENTRY_EFFECT",
    "HOT_RANK_CHANGED",
    "HOT_RANK_CHANGED_V2",
    # 'INTERACT_WORD', # 进入直播间
    # 'LIVE',
    "LIVE_INTERACTIVE_GAME",
    "NOTICE_MSG",
    "ONLINE_RANK_COUNT",
    "ONLINE_RANK_TOP3",
    "ONLINE_RANK_V2",
    "PK_BATTLE_END",
    "PK_BATTLE_ENTRANCE",
    "PK_BATTLE_FINAL_PROCESS",
    "PK_BATTLE_MATCH_TIMEOUT",
    "PK_BATTLE_PRE",
    "PK_BATTLE_PRE_NEW",
    "PK_BATTLE_PROCESS",
    "PK_BATTLE_PROCESS_NEW",
    "PK_BATTLE_SETTLE",
    "PK_BATTLE_SETTLE_NEW",
    "PK_BATTLE_SETTLE_USER",
    "PK_BATTLE_SETTLE_V2",
    "PK_BATTLE_START",
    "PK_BATTLE_START_NEW",
    "PK_BATTLE_VIDEO_PUNISH_BEGIN",
    "PK_BATTLE_VIDEO_PUNISH_END",
    # 'PREPARING',
    "ROOM_REAL_TIME_MESSAGE_UPDATE",
    "STOP_LIVE_ROOM_LIST",
    "SUPER_CHAT_MESSAGE_JPN",
    "WIDGET_BANNER",
    "WATCHED_CHANGE",
    "LIVE_MULTI_VIEW_CHANGE",
    "COMMON_NOTICE_DANMAKU",  # BLS秋季限时任务：任务即将结束，抓紧完成获取10元红包奖励吧！未完成任务进度将无法获得奖励 https://st.openbayes.net/organizations/openbayes/issues/732/events/1fcc24e45ef34148b3193fb676dffc5a/?project=13&query=is%3Aunresolved
    "DANMU_AGGREGATION",  # 老板大气！点点红包抽礼物, https://st.openbayes.net/organizations/openbayes/issues/732/events/aafcb5e6c02b45af8de99a26175e25fe/?project=13&query=is%3Aunresolved
    "HOT_RANK_SETTLEMENT",  # 恭喜主播 <% 阿梓从小就很可爱 %> 荣登限时热门榜虚拟主播榜top10! 即将获得热门流量推荐哦！
    "HOT_RANK_SETTLEMENT_V2",  # 恭喜主播 <% 老实憨厚的笑笑 %> 荣登限时热门榜户外榜榜首! 即将获得热门流量推荐哦！
}
"""已打日志的未知cmd"""


class HandlerInterface:
    """
    直播消息处理器接口
    """

    def handle(self, client: ws_base.WebSocketClientBase, command: dict):
        raise NotImplementedError

    def on_client_stopped(
        self, client: ws_base.WebSocketClientBase, exception: Optional[Exception]
    ):
        """
        当客户端停止时调用。可以在这里close或者重新start
        """


def _make_msg_callback(method_name, message_cls):
    def callback(
        self: "BaseHandler", client: ws_base.WebSocketClientBase, command: dict
    ):
        method = getattr(self, method_name)
        return method(client, message_cls.from_command(command["data"]))

    return callback


def _make_raw_msg_callback(method_name, message_cls, raw=False):
    def callback(
        self: "BaseHandler", client: ws_base.WebSocketClientBase, command: dict
    ):
        method = getattr(self, method_name)
        return method(
            client, message_cls.from_command(command if raw else command["data"])
        )

    return callback


class BaseHandler(HandlerInterface):
    """
    一个简单的消息处理器实现，带消息分发和消息类型转换。继承并重写_on_xxx方法即可实现自己的处理器
    """

    def __danmu_msg_callback(self, client: ws_base.WebSocketClientBase, command: dict):
        return self._on_danmaku(
            client, web_models.DanmakuMessage.from_command(command["info"])
        )

    _CMD_CALLBACK_DICT: Dict[
        str, Optional[Callable[["BaseHandler", ws_base.WebSocketClientBase, dict], Any]]
    ] = {
        # 收到心跳包，这是blivedm自造的消息，原本的心跳包格式不一样
        "_HEARTBEAT": _make_msg_callback("_on_heartbeat", web_models.HeartbeatMessage),
        # 收到弹幕
        # go-common\app\service\live\live-dm\service\v1\send.go
        "DANMU_MSG": __danmu_msg_callback,
        # 有人送礼
        "SEND_GIFT": _make_msg_callback("_on_gift", web_models.GiftMessage),
        # 有人上舰
        "GUARD_BUY": _make_msg_callback("_on_buy_guard", web_models.GuardBuyMessage),
        # 醒目留言
        "SUPER_CHAT_MESSAGE": _make_msg_callback(
            "_on_super_chat", web_models.SuperChatMessage
        ),
        # 删除醒目留言
        "SUPER_CHAT_MESSAGE_DELETE": _make_msg_callback(
            "_on_super_chat_delete", web_models.SuperChatDeleteMessage
        ),
        # eop-blive 自定义事件
        "LIVE": _make_raw_msg_callback("_on_live", web_models.LiveMessage, True),
        "PREPARING": _make_raw_msg_callback(
            "_on_preparing", web_models.PreparingMessage, True
        ),
        "INTERACT_WORD": _make_raw_msg_callback(
            "_on_interact_word", web_models.InteractWordMessage
        ),
        "USER_TOAST_MSG": _make_raw_msg_callback(
            "_on_user_toast", web_models.UserToastMessage
        ),
        "USER_VIRTUAL_MVP": _make_raw_msg_callback(
            "_on_virtual_mvp", web_models.UserVirtualMvp
        ),
        "WARNING": _make_raw_msg_callback(
            "_on_warning", web_models.WarningMessage, True
        ),
        "CUT_OFF": _make_raw_msg_callback(
            "_on_cut_off", web_models.CutOffMessage, True
        ),
        "ROOM_SILENT_ON": _make_raw_msg_callback(
            "_on_room_silent_on", web_models.RoomSilentOnMessage
        ),
        "ROOM_SILENT_OFF": _make_raw_msg_callback(
            "_on_room_silent_off", web_models.RoomSilentOffMessage
        ),
        "ROOM_ADMINS": _make_raw_msg_callback(
            "_on_room_admins", web_models.RoomAdminsMessage, True
        ),
        # yes this is the lowercased event from ws 😅
        "room_admin_entrance": _make_raw_msg_callback(
            "_on_room_admin_entrance", web_models.RoomAdminEntranceMessage, True
        ),
        "ROOM_ADMIN_REVOKE": _make_raw_msg_callback(
            "_on_room_admin_revoke", web_models.RoomAdminRevokeMessage, True
        ),
        "ROOM_BLOCK_MSG": _make_raw_msg_callback(
            "_on_room_block", web_models.RoomBlockMessage
        ),
        #
        # 开放平台消息
        #
        # 收到弹幕
        "LIVE_OPEN_PLATFORM_DM": _make_msg_callback(
            "_on_open_live_danmaku", open_models.DanmakuMessage
        ),
        # 有人送礼
        "LIVE_OPEN_PLATFORM_SEND_GIFT": _make_msg_callback(
            "_on_open_live_gift", open_models.GiftMessage
        ),
        # 有人上舰
        "LIVE_OPEN_PLATFORM_GUARD": _make_msg_callback(
            "_on_open_live_buy_guard", open_models.GuardBuyMessage
        ),
        # 醒目留言
        "LIVE_OPEN_PLATFORM_SUPER_CHAT": _make_msg_callback(
            "_on_open_live_super_chat", open_models.SuperChatMessage
        ),
        # 删除醒目留言
        "LIVE_OPEN_PLATFORM_SUPER_CHAT_DEL": _make_msg_callback(
            "_on_open_live_super_chat_delete", open_models.SuperChatDeleteMessage
        ),
        # 点赞
        "LIVE_OPEN_PLATFORM_LIKE": _make_msg_callback(
            "_on_open_live_like", open_models.LikeMessage
        ),
    }  # type: ignore
    """cmd -> 处理回调"""

    def handle(self, client: ws_base.WebSocketClientBase, command: dict):
        cmd = command.get("cmd", "")
        pos = cmd.find(":")  # 2019-5-29 B站弹幕升级新增了参数
        if pos != -1:
            cmd = cmd[:pos]

        if cmd not in self._CMD_CALLBACK_DICT:
            # 只有第一次遇到未知cmd时打日志
            if cmd not in logged_unknown_cmds:
                logger.warning(
                    "room=%d unknown cmd=%s, command=%s", client.room_id, cmd, command
                )
                logged_unknown_cmds.add(cmd)
            return

        callback = self._CMD_CALLBACK_DICT[cmd]
        if callback is not None:
            callback(self, client, command)

    def _on_heartbeat(
        self, client: ws_base.WebSocketClientBase, message: web_models.HeartbeatMessage
    ):
        """
        收到心跳包
        """

    def _on_danmaku(
        self, client: ws_base.WebSocketClientBase, message: web_models.DanmakuMessage
    ):
        """
        收到弹幕
        """

    def _on_gift(
        self, client: ws_base.WebSocketClientBase, message: web_models.GiftMessage
    ):
        """
        收到礼物
        """

    def _on_buy_guard(
        self, client: ws_base.WebSocketClientBase, message: web_models.GuardBuyMessage
    ):
        """
        有人上舰
        """

    def _on_super_chat(
        self, client: ws_base.WebSocketClientBase, message: web_models.SuperChatMessage
    ):
        """
        醒目留言
        """

    def _on_super_chat_delete(
        self,
        client: ws_base.WebSocketClientBase,
        message: web_models.SuperChatDeleteMessage,
    ):
        """
        删除醒目留言
        """

    # eop-blive 自定义事件
    def _on_live(
        self,
        client: ws_base.WebSocketClientBase,
        message: web_models.LiveMessage,
    ):
        """
        _on_live
        """

    def _on_preparing(
        self,
        client: ws_base.WebSocketClientBase,
        message: web_models.PreparingMessage,
    ):
        """
        _on_preparing
        """

    def _on_interact_word(
        self,
        client: ws_base.WebSocketClientBase,
        message: web_models.InteractWordMessage,
    ):
        """
        _on_interact_word
        """

    def _on_user_toast(
        self, client: ws_base.WebSocketClientBase, message: web_models.UserToastMessage
    ):
        """
        _on_user_toast
        """

    def _on_virtual_mvp(
        self, client: ws_base.WebSocketClientBase, message: web_models.UserVirtualMvp
    ):
        """
        _on_virtual_mvp
        """

    def _on_warning(
        self, client: ws_base.WebSocketClientBase, message: web_models.WarningMessage
    ):
        """
        _on_warning
        """

    def _on_cut_off(
        self, client: ws_base.WebSocketClientBase, message: web_models.CutOffMessage
    ):
        """
        _on_cut_off
        """

    def _on_room_silent_on(
        self,
        client: ws_base.WebSocketClientBase,
        message: web_models.RoomSilentOnMessage,
    ):
        """
        _on_room_silent_on
        """

    def _on_room_silent_off(
        self,
        client: ws_base.WebSocketClientBase,
        message: web_models.RoomSilentOffMessage,
    ):
        """
        _on_room_silent_off
        """

    def _on_room_admins(
        self, client: ws_base.WebSocketClientBase, message: web_models.RoomAdminsMessage
    ):
        """
        _on_room_admins
        """

    def _on_room_admin_entrance(
        self,
        client: ws_base.WebSocketClientBase,
        message: web_models.RoomAdminEntranceMessage,
    ):
        """
        _on_room_admin_entrance
        """

    def _on_room_admin_revoke(
        self,
        client: ws_base.WebSocketClientBase,
        message: web_models.RoomAdminRevokeMessage,
    ):
        """
        _on_room_admin_revoke
        """

    def _on_room_block(
        self, client: ws_base.WebSocketClientBase, message: web_models.RoomBlockMessage
    ):
        """
        _on_room_block
        """

    #
    # 开放平台消息
    #

    def _on_open_live_danmaku(
        self, client: ws_base.WebSocketClientBase, message: open_models.DanmakuMessage
    ):
        """
        收到弹幕
        """

    def _on_open_live_gift(
        self, client: ws_base.WebSocketClientBase, message: open_models.GiftMessage
    ):
        """
        收到礼物
        """

    def _on_open_live_buy_guard(
        self, client: ws_base.WebSocketClientBase, message: open_models.GuardBuyMessage
    ):
        """
        有人上舰
        """

    def _on_open_live_super_chat(
        self, client: ws_base.WebSocketClientBase, message: open_models.SuperChatMessage
    ):
        """
        醒目留言
        """

    def _on_open_live_super_chat_delete(
        self,
        client: ws_base.WebSocketClientBase,
        message: open_models.SuperChatDeleteMessage,
    ):
        """
        删除醒目留言
        """

    def _on_open_live_like(
        self, client: ws_base.WebSocketClientBase, message: open_models.LikeMessage
    ):
        """
        点赞
        """
